import torch
from torchvision import transforms, models
from model import get_train_loss_needs, get_plant_model
from dataset import get_dataloaders, get_datasets
from classification import train_model, save_learning_curve

data_transforms = {
    'train': transforms.Compose(
        [
            transforms.RandomResizedCrop(224),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
        ]
    ),
    'test': transforms.Compose(
        [
            transforms.Resize(256),
            transforms.CenterCrop(224),
            transforms.ToTensor(),
            transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
        ]
    ),
}

if __name__ == "__main__":
    data_dir = '/Users/angelmancera/Columbia/Classes/Spring_2023/Open_Src_Dev/Taiwan_Tomato_Leaves'
    image_datasets = get_datasets(data_dir, data_transforms)
    dataloaders = get_dataloaders(image_datasets)

    dataset_sizes = {x: len(image_datasets[x]) for x in ['train', 'test']}
    class_names = image_datasets['train'].classes

    # Sanity Checks
    # print(dataset_sizes)
    # print(class_names)

    # Configure to train with GPU
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")

    # Load pre-trained model and set add classification layer
    model = models.resnet18(pretrained=True)
    model_ft = get_plant_model(model, class_names)
    model_ft = model_ft.to(device)

    # Set up loss, optimizer, and scheduler
    criterion, optimizer_ft, exp_lr_scheduler = get_train_loss_needs(
        model_ft, lr=0.001, momentum=0.9, step_size=7, gamma=0.1
    )

    # Train and evaluate model
    model_ft, train_losses, train_accuracies, val_losses, val_accuracies = train_model(
        device, dataset_sizes, dataloaders, model_ft, criterion, optimizer_ft, exp_lr_scheduler, num_epochs=16
    )

    # Let's save our learning curve as well
    save_learning_curve(train_losses, train_accuracies, val_losses, val_accuracies)



import torch
import numpy as np
import matplotlib.pyplot as plt
import time
import copy


def train_model(
    device, dataset_sizes, dataloaders, model, criterion, optimizer, scheduler, num_epochs=25, testing=False
):
    since = time.time()

    best_model_wts = copy.deepcopy(model.state_dict())
    best_acc = 0.0

    train_losses = []
    val_losses = []
    train_accuracies = []
    val_accuracies = []

    for epoch in range(num_epochs):
        print(f'Epoch {epoch}/{num_epochs - 1}')
        print('-' * 10)

        for phase in ['train', 'test']:
            if phase == 'train':
                model.train()
            else:
                model.eval()

            running_loss = 0.0
            running_corrects = 0

            # Iterate over data.
            for inputs, labels in dataloaders[phase]:
                inputs = inputs.to(device)
                labels = labels.to(device)

                # zero the parameter gradients
                optimizer.zero_grad()

                # forward
                # track history if only in train
                with torch.set_grad_enabled(phase == 'train'):
                    outputs = model(inputs)
                    _, preds = torch.max(outputs, 1)
                    loss = criterion(outputs, labels)

                    # backward + optimize only if in training phase
                    if phase == 'train':
                        loss.backward()
                        optimizer.step()

                # statistics
                running_loss += loss.item() * inputs.size(0)
                running_corrects += torch.sum(preds == labels.data)
            if phase == 'train':
                scheduler.step()

            epoch_loss = running_loss / dataset_sizes[phase]
            epoch_acc = running_corrects.double() / dataset_sizes[phase]
            if phase == "train":
                train_losses.append(epoch_loss)
                train_accuracies.append(epoch_acc)
            else:
                val_losses.append(epoch_loss)
                val_accuracies.append(epoch_acc)

            print(f'{phase} Loss: {epoch_loss:.4f} Acc: {epoch_acc:.4f}')

            if phase == 'test' and epoch_acc > best_acc:
                best_acc = epoch_acc
                best_model_wts = copy.deepcopy(model.state_dict())

            if testing:
                break

        print()

    time_elapsed = time.time() - since
    print(f'Training complete in {time_elapsed // 60:.0f}m {time_elapsed % 60:.0f}s')
    print(f'Best val Acc: {best_acc:4f}')

    # load best model weights
    model.load_state_dict(best_model_wts)
    return model, train_losses, train_accuracies, val_losses, val_accuracies


def save_learning_curve(train_loss_list, train_miou_list, val_loss_list, val_miou_list):
    epochs = np.arange(1, len(train_loss_list) + 1)
    plt.figure()
    plt.plot(epochs, train_loss_list, color='navy', label="train_loss")
    plt.plot(epochs, train_miou_list, color='teal', label="train_mIoU")
    plt.plot(epochs, val_loss_list, color='orange', label="val_loss")
    plt.plot(epochs, val_miou_list, color='gold', label="val_mIoU")
    plt.legend(bbox_to_anchor=(1.05, 1), loc='upper left', borderaxespad=0.0)
    plt.xticks(epochs, epochs)
    plt.yticks(np.arange(10) * 0.1, [f"0.{i}" for i in range(10)])
    plt.xlabel('epoch')
    plt.ylabel('mIoU')
    plt.grid(True)
    plt.savefig('learning_curve.png', bbox_inches='tight')
    plt.show()



import torch.nn as nn
import torch.optim as optim
from torch.optim import lr_scheduler


def get_plant_model(model, class_names):
    num_ftrs = model.fc.in_features
    model.fc = nn.Linear(num_ftrs, len(class_names))
    return model


def get_train_loss_needs(model_ft, lr, momentum, step_size, gamma):
    criterion = nn.CrossEntropyLoss()
    optimizer_ft = optim.SGD(model_ft.parameters(), lr=lr, momentum=momentum)
    exp_lr_scheduler = lr_scheduler.StepLR(optimizer_ft, step_size=step_size, gamma=gamma)
    return criterion, optimizer_ft, exp_lr_scheduler



from .classification import train_model, save_learning_curve
from .model import get_plant_model, get_train_loss_needs
from .dataset import PlantDataset, collate_fn, get_datasets, get_dataloaders

__version__ = "0.1.1"



import os
import glob
import torch
import itertools
import numpy as np
from PIL import Image


class PlantDataset(torch.utils.data.Dataset):
    def __init__(self, images_dir, image_transform):
        self.image_transform = image_transform

        # Get a list of all class names inside the images directory
        self.classes = self.get_class_names(images_dir)

        # Assign a unique label index to each class name
        self.class_labels = {name: idx for idx, name in enumerate(self.classes)}

        image_files, labels = self.get_image_filenames_with_labels(
            images_dir,
            self.classes,
            self.class_labels,
        )

        # This is a trick to avoid memory leaks over very large datasets.
        self.image_files = np.array(image_files)
        self.labels = np.array(labels).astype("int")

        # How many total images do we need to iterate in this entire dataset?
        self.num_images = len(self.image_files)

    def __len__(self):
        return self.num_images

    def get_class_names(self, images_dir):
        """
        Given a directory of images, underneath which we have directories of class
        names, collect all these class names and return as a list of strings.
        """
        class_name_dirs = glob.glob(images_dir + "/*")
        class_names = [name.replace(images_dir + "/", "") for name in class_name_dirs]
        return sorted(class_names)  # sort just to keep consistency

    def get_image_filenames_with_labels(self, images_dir, class_names, class_labels):
        image_files = []
        labels = []

        supported_file_types = ["/*.jpg", "/*.jpeg", "/*.gif"]

        for name in class_names:
            # Glob all (supported) image file names in this directory
            image_class_dir = os.path.join(images_dir, name)

            # Iterate through the supported file types.  For each, glob a list of
            # all file names with that file extension.  Then combine the entire list
            # into one list using itertools.
            image_class_files = list(
                itertools.chain.from_iterable(
                    [glob.glob(image_class_dir + file_type) for file_type in supported_file_types]
                )
            )

            # Concatenate the image file names to the overall list and create a label for each
            image_files += image_class_files
            labels += [class_labels[name]] * len(image_class_files)

        return image_files, labels

    def __getitem__(self, idx):
        try:
            image = Image.open(self.image_files[idx]).convert('RGB')
            label = self.labels[idx]

            # Apply the image transform
            image = self.image_transform(image)

            return image, label
        except Exception as exc:
            return exc


def collate_fn(batch):
    # Filter failed images first
    batch = list(filter(lambda x: x is not None, batch))

    # Now collate into mini-batches
    images = torch.stack([b[0] for b in batch])
    labels = torch.LongTensor([b[1] for b in batch])

    return images, labels


def get_datasets(data_dir, data_transforms):
    return {x: PlantDataset(os.path.join(data_dir, x), data_transforms[x]) for x in ['train', 'test']}


def get_dataloaders(image_datasets):
    return {
        x: torch.utils.data.DataLoader(
            image_datasets[x], batch_size=4, shuffle=True, num_workers=0, collate_fn=collate_fn
        )
        for x in ['train', 'test']
    }



from multi_scale_expansion import *
import unittest
from unittest.mock import patch
import torch
import torch.nn as nn
from torchvision import transforms, models
from torchvision.datasets import FakeData


## Tests related to our Dataset
class TestDataset(unittest.TestCase):
    def test_collate_fn(self):
        first_tensor = torch.randn((576, 576, 3))
        second_tensor = torch.randn((576, 576, 3))
        batch = [[first_tensor, 0], [second_tensor, 1]]
        output = collate_fn(batch)

        # Expected
        images = torch.stack([batch[0][0], batch[1][0]])
        labels = torch.LongTensor([batch[0][1], batch[1][1]])

        assert torch.equal(output[0], images)
        assert torch.equal(output[1], labels)

    def test_get_datasets_fn(self):  # mock_dir, mock_transforms
        mock_dir = "/Users/angelmancera/Columbia/Classes/Spring_2023/Open_Src_Dev/Taiwan_Tomato_Leaves"
        mock_transforms = {
            'train': transforms.Compose(
                [
                    transforms.RandomResizedCrop(224),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
                ]
            ),
            'test': transforms.Compose(
                [
                    transforms.Resize(256),
                    transforms.CenterCrop(224),
                    transforms.ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
                ]
            ),
        }
        datasets = get_datasets(mock_dir, mock_transforms)

        for ds in datasets.items():
            assert ds[0] in ["train", "test"]
            assert isinstance(ds[1], torch.utils.data.Dataset)

    def test_get_dataloaders_fn(self):  # mock_datasets
        data_transforms = {
            'train': transforms.Compose(
                [
                    transforms.RandomResizedCrop(224),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
                ]
            ),
            'test': transforms.Compose(
                [
                    transforms.Resize(256),
                    transforms.CenterCrop(224),
                    transforms.ToTensor(),
                    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
                ]
            ),
        }
        mock_datasets = {
            "train": FakeData(num_classes=6, transform=data_transforms["train"]),
            "test": FakeData(num_classes=6, transform=data_transforms["test"]),
        }
        dataloaders = get_dataloaders(mock_datasets)
        for dl in dataloaders.values():
            assert isinstance(dl, torch.utils.data.DataLoader)


## Tests related to our model
def test_get_plant_model():  # mock_model
    mock_model = models.resnet18(weights='DEFAULT')
    class_names = list(range(6))
    model = get_plant_model(mock_model, class_names)
    assert isinstance(model, nn.Module)


def test_get_train_loss_needs():  # mock_model, mock_lr, mock_momentum, mock_step_size, mock_gamma
    mock_model = models.resnet18(weights='DEFAULT')
    mock_lr = 0.001
    mock_momentum = 0.9
    mock_step_size = 7
    mock_gamma = 0.1
    needs = get_train_loss_needs(mock_model, mock_lr, mock_momentum, mock_step_size, mock_gamma)
    criterion = needs[0]
    assert isinstance(criterion, nn.CrossEntropyLoss)
    optimizer = needs[1]
    assert isinstance(optimizer, torch.optim.SGD)
    lr_scheduler = needs[2]
    assert isinstance(lr_scheduler, torch.optim.lr_scheduler.StepLR)


## Tests related to our classification training and visualization
# Integration test
def test_train_model():  # mock_dir, mock_transforms
    device = torch.device("cuda:0" if torch.cuda.is_available() else "cpu")
    mock_model = models.resnet18(weights='DEFAULT')
    class_names = list(range(6))
    mock_model = get_plant_model(mock_model, class_names)

    mock_model = models.resnet18(weights='DEFAULT')
    mock_lr = 0.001
    mock_momentum = 0.9
    mock_step_size = 7
    mock_gamma = 0.1
    mock_criterion, mock_optimizer, mock_lr_scheduler = get_train_loss_needs(
        mock_model, mock_lr, mock_momentum, mock_step_size, mock_gamma
    )

    mock_transforms = {
        'train': transforms.Compose(
            [
                transforms.RandomResizedCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        ),
        'test': transforms.Compose(
            [
                transforms.Resize(256),
                transforms.CenterCrop(224),
                transforms.ToTensor(),
                transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225]),
            ]
        ),
    }
    mock_datasets = {
        "train": FakeData(num_classes=6, transform=mock_transforms["train"]),
        "test": FakeData(num_classes=6, transform=mock_transforms["test"]),
    }
    mock_dataset_sizes = {x: len(mock_datasets[x]) for x in ['train', 'test']}
    mock_dataloaders = get_dataloaders(mock_datasets)

    model, train_losses, train_accuracies, val_losses, val_accuracies = train_model(
        device,
        mock_dataset_sizes,
        mock_dataloaders,
        mock_model,
        mock_criterion,
        mock_optimizer,
        mock_lr_scheduler,
        num_epochs=1,
        testing=True,
    )
    assert isinstance(model, nn.Module)
    assert isinstance(train_losses, list)
    assert isinstance(train_accuracies, list)
    assert isinstance(val_losses, list)
    assert isinstance(val_accuracies, list)
